#include <stdio.h>
#include <stdlib.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "type.h"
#include "asm_symbol.h"
#include "asm_format.h"
#include "asm_code.h"
#include "asm_code_aarch64.h"

static int asm_code_aarch64_get_word_size(void)
{
  return 8;
}

static int asm_code_aarch64_get_pointer_size(void)
{
  return asm_code_aarch64_get_word_size();
}

static int asm_code_aarch64_funccall_args_reg_number(void)
{
  return 4; /* r0/r1/r2/r3 (See funccall_regs[]) */
}

static int asm_code_aarch64_funccall_args_stack_number(void)
{
  return VALUE_FUNCTION_ARGS_NUM;
}

static int asm_code_aarch64_tmp_reg_number(void)
{
  return 4; /* x16/x17/x18/x19 (See tmp_regs[]) */
}

static int asm_code_aarch64_tmp_stack_number(void)
{
  return VALUE_FUNCTION_ARGS_NUM;
}

static int asm_code_aarch64_function_register_number(void)
{
  return 4; /* x16/x17/x18/x19 (See asm_code_aarch64_function_register_save() */
}

static int asm_code_aarch64_function_saveparam_number(void)
{
  return 2; /* x30(LR)/x29(FP) (See asm_code_aarch64_function_start() */
}

static int asm_code_aarch64_stack_align_size(void)
{
  return 16;
}

static int asm_code_aarch64_stack_correct_size(void)
{
  return 0;
}

static char *r0_by_size(model_t model)
{
  if (model != NULL) {
    switch (model_get_size(model)) {
    case -1: ERREXIT(1);
    case  1: return "w0";
    case  2: return "w0";
    case  4: return "w0";
    default: break;
    }
  }
  return "x0";
}

static char *r1_by_size(model_t model)
{
  if (model != NULL) {
    switch (model_get_size(model)) {
    case -1: ERREXIT(1);
    case  1: return "w1";
    case  2: return "w1";
    case  4: return "w1";
    default: break;
    }
  }
  return "x1";
}

static char *load_by_size(model_t model)
{
  if (model != NULL) {
    switch (model_get_size(model)) {
    case -1: ERREXIT(1);
    case  1: return "ldrb";
    case  2: return "ldrh";
    case  4: return "ldr";
    default: break;
    }
  }
  return "ldr";
}

static char *store_by_size(model_t model)
{
  if (model != NULL) {
    switch (model_get_size(model)) {
    case -1: ERREXIT(1);
    case  1: return "strb";
    case  2: return "strh";
    case  4: return "str";
    default: break;
    }
  }
  return "str";
}

static void asm_code_aarch64_get_value(FILE *out, long value)
{
  ASM_CODE_OUT(out, "\tmov\tmov\tx0, #0x%lx", value & 0xffff);
  ASM_CODE_OUT(out, "\tmov\tmovk\tx0, #0x%lx, lsl #16", (value >> 16) & 0xffff);
}

static void asm_code_aarch64_get_value_r1(FILE *out, long value)
{
  ASM_CODE_OUT(out, "\tmov\tmov\tx1, #0x%lx", value & 0xffff);
  ASM_CODE_OUT(out, "\tmov\tmovk\tx1, #0x%lx, lsl #16", (value >> 16) & 0xffff);
}

static void asm_code_aarch64_get_address_stack(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\tx0, sp\n");
}

static void asm_code_aarch64_get_address(FILE *out, char *label)
{
  asm_symbol_tmp_label_reserve(2);
  ASM_CODE_OUT(out, "\tldr\tx0, %s\n", asm_symbol_tmp_label_namef(0));
  ASM_CODE_OUT(out, "\tb\t%s\n", asm_symbol_tmp_label_namef(1));
  asm_format_label(out, asm_symbol_tmp_label_name(0));
  ASM_CODE_OUT(out, "\t.wordx\t%s\n", label);
  asm_format_label(out, asm_symbol_tmp_label_name(1));
  asm_symbol_tmp_label_release(2);
}

static void asm_code_aarch64_add_address(FILE *out, int offset)
{
  ASM_CODE_OUT(out, "\tadd\tx0, x0, #0x%x\n", offset);
}

static void asm_code_aarch64_get_r1(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\tx0, x1\n");
}

static void asm_code_aarch64_set_r1(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\tx1, x0\n");
}

static void asm_code_aarch64_memory_load(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\t%s, [x1, #0x%x]\n",
	       load_by_size(model), r0_by_size(model), offset);
}

static void asm_code_aarch64_memory_store(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\t%s, [x1, #0x%x]\n",
	       store_by_size(model), r0_by_size(model), offset);
}

static void asm_code_aarch64_stack_load(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\t%s, [sp, #0x%x]\n",
	       load_by_size(model), r0_by_size(model), offset);
}

static void asm_code_aarch64_stack_store(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\t%s, [sp, #0x%x]\n",
	       store_by_size(model), r0_by_size(model), offset);
}

static void asm_code_aarch64_stack_load_r1(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\t%s, [sp, #0x%x]\n",
	       load_by_size(model), r1_by_size(model), offset);
}

static void asm_code_aarch64_stack_store_r1(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\t%s, [sp, #0x%x]\n",
	       store_by_size(model), r1_by_size(model), offset);
}

static void asm_code_aarch64_stack_expand(FILE *out, int size)
{
  ASM_CODE_OUT(out, "\tsub\tsp, sp, #0x%x\n", size);
}

static void asm_code_aarch64_stack_reduce(FILE *out, int size)
{
  ASM_CODE_OUT(out, "\tadd\tsp, sp, #0x%x\n", size);
}

static char *funccall_regs[] = {
  "x0", "x1", "x2", "x3", "x4", "x5", "x6", "x7"
};

static void asm_code_aarch64_funccall_reg_load(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\tx0, %s\n", funccall_regs[number]);
}

static void asm_code_aarch64_funccall_reg_store(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%s, x0\n", funccall_regs[number]);
}

static char *tmp_regs[] = {
  "x16", "x17", "x18", "x19",
};

static void asm_code_aarch64_tmp_reg_load(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\tx0, %s\n", tmp_regs[number]);
}

static void asm_code_aarch64_tmp_reg_save(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%s, x0\n", tmp_regs[number]);
}

static void asm_code_aarch64_tmp_reg_load_r1(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\tx1, %s\n", tmp_regs[number]);
}

static void asm_code_aarch64_tmp_reg_save_r1(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%s, x1\n", tmp_regs[number]);
}

static int asm_code_aarch64_sign_extension_char(FILE *out)
{
  ASM_CODE_OUT(out, "\tsxtb\tx0, w0\n");
  return 0;
}

static int asm_code_aarch64_sign_extension_uchar(FILE *out)
{
  ASM_CODE_OUT(out, "\tuxtb\tw0, w0\n");
  return 0;
}

static int asm_code_aarch64_sign_extension_short(FILE *out)
{
  ASM_CODE_OUT(out, "\tsxth\tx0, w0\n");
  return 0;
}

static int asm_code_aarch64_sign_extension_ushort(FILE *out)
{
  ASM_CODE_OUT(out, "\tuxth\tw0, w0\n");
  return 0;
}

static int asm_code_aarch64_sign_extension_int(FILE *out)
{
  ASM_CODE_OUT(out, "\tsxtw\tx0, w0\n");
  return 0;
}

static int asm_code_aarch64_sign_extension_uint(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\tw0, w0\n");
  return 0;
}

static int asm_code_aarch64_calc_inv(FILE *out, model_t model, model_t model_arg0)
{
  ASM_CODE_OUT(out, "\tmvn\tx0, x0\n");
  return 0;
}

static int asm_code_aarch64_calc_minus(FILE *out, model_t model, model_t model_arg0)
{
  ASM_CODE_OUT(out, "\tneg\tx0, x0\n");
  return 0;
}

static int asm_code_aarch64_calc_op1(FILE *out, c_type_t type, model_t model, model_t model_arg0)
{
  return -1;
}

static void asm_code_aarch64_calc_add(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tadd\tx0, x0, x1\n");
}

static void asm_code_aarch64_calc_sub(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tsub\tx0, x0, x1\n");
}

static void asm_code_aarch64_calc_and(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tand\tx0, x0, x1\n");
}

static void asm_code_aarch64_calc_or(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\torr\tx0, x0, x1\n");
}

static void asm_code_aarch64_calc_xor(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\teor\tx0, x0, x1\n");
}

static int asm_code_aarch64_calc_mul(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tmul\tx0, x0, x1\n");
  return 0;
}

static int asm_code_aarch64_calc_div(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tsdiv\tx0, x0, x1\n");
  return 0;
}

static int asm_code_aarch64_calc_mod(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tsdiv\tx2, x0, x1\n");
  ASM_CODE_OUT(out, "\tmsub\tx0, x2, x1, x0\n");
  return 0;
}

static int asm_code_aarch64_calc_llshift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tlsl\tx0, x0, x1\n");
  return 0;
}

static int asm_code_aarch64_calc_rashift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tasr\tx0, x0, x1\n");
  return 0;
}

static int asm_code_aarch64_calc_rlshift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tlsr\tx0, x0, x1\n");
  return 0;
}

static int asm_code_aarch64_calc_op2(FILE *out, c_type_t type, model_t model, model_t model_arg0, model_t model_arg1)
{
  return -1;
}

static void asm_code_aarch64_branch(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tb\t%s\n", label);
}

static int asm_code_aarch64_branch_zero(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tcmp\tx0, xzr\n");
  ASM_CODE_OUT(out, "\tb.eq\t%s\n", label);
  return 0;
}

static int asm_code_aarch64_branch_nzero(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tcmp\tx0, xzr\n");
  ASM_CODE_OUT(out, "\tb.nq\t%s\n", label);
  return 0;
}

static int asm_code_aarch64_branch_cmp_eq(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tx0, x1\n");
  ASM_CODE_OUT(out, "\tb.eq\t%s\n", label);
  return 0;
}

static int asm_code_aarch64_branch_cmp_ne(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tx0, x1\n");
  ASM_CODE_OUT(out, "\tb.ne\t%s\n", label);
  return 0;
}

static int asm_code_aarch64_branch_cmp_lt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tx0, x1\n");
  ASM_CODE_OUT(out, "\tb.lt\t%s\n", label);
  return 0;
}

static int asm_code_aarch64_branch_cmp_gt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tx0, x1\n");
  ASM_CODE_OUT(out, "\tb.gt\t%s\n", label);
  return 0;
}

static int asm_code_aarch64_branch_cmp_le(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tx0, x1\n");
  ASM_CODE_OUT(out, "\tb.le\t%s\n", label);
  return 0;
}

static int asm_code_aarch64_branch_cmp_ge(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tx0, x1\n");
  ASM_CODE_OUT(out, "\tb.ge\t%s\n", label);
  return 0;
}

static int asm_code_aarch64_branch_cmp_ult(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tx0, x1\n");
  ASM_CODE_OUT(out, "\tb.cc\t%s\n", label);
  return 0;
}

static int asm_code_aarch64_branch_cmp_ugt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tx0, x1\n");
  ASM_CODE_OUT(out, "\tb.hi\t%s\n", label);
  return 0;
}

static int asm_code_aarch64_branch_cmp_ule(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tx0, x1\n");
  ASM_CODE_OUT(out, "\tb.ls\t%s\n", label);
  return 0;
}

static int asm_code_aarch64_branch_cmp_uge(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tx0, x1\n");
  ASM_CODE_OUT(out, "\tb.cs\t%s\n", label);
  return 0;
}

static int asm_code_aarch64_branch_cmp(FILE *out, char *label, c_type_t type, model_t model)
{
  return -1;
}

static void asm_code_aarch64_function_call(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tbl\t%s\n", label);
}

static void asm_code_aarch64_function_call_pointer(FILE *out)
{
  ASM_CODE_OUT(out, "\tblr\tx0\n");
}

static void asm_code_aarch64_function_start(FILE *out)
{
  ASM_CODE_OUT(out, "\tsub\tsp, sp, #0x10\n");
  ASM_CODE_OUT(out, "\tstr\tx30, [sp, #0x8]\n");
  ASM_CODE_OUT(out, "\tstr\tx29, [sp]\n");
}

static void asm_code_aarch64_function_end(FILE *out)
{
  ASM_CODE_OUT(out, "\tldr\tx29, [sp]\n");
  ASM_CODE_OUT(out, "\tldr\tx30, [sp, #0x8]\n");
  ASM_CODE_OUT(out, "\tadd\tsp, sp, #0x10\n");
  ASM_CODE_OUT(out, "\tret\n");
}

static void asm_code_aarch64_function_register_save(FILE *out)
{
  ASM_CODE_OUT(out, "\tsub\tsp, sp, #0x20\n");
  ASM_CODE_OUT(out, "\tstr\tx19, [sp, #0x18]\n");
  ASM_CODE_OUT(out, "\tstr\tx18, [sp, #0x10]\n");
  ASM_CODE_OUT(out, "\tstr\tx17, [sp, #0x8]\n");
  ASM_CODE_OUT(out, "\tstr\tx16, [sp]\n");
}

static void asm_code_aarch64_function_register_load(FILE *out)
{
  ASM_CODE_OUT(out, "\tldr\tx16, [sp]\n");
  ASM_CODE_OUT(out, "\tldr\tx17, [sp, #0x8]\n");
  ASM_CODE_OUT(out, "\tldr\tx18, [sp, #0x10]\n");
  ASM_CODE_OUT(out, "\tldr\tx19, [sp, #0x18]\n");
  ASM_CODE_OUT(out, "\tadd\tsp, sp, #0x20\n");
}

struct asm_code_callback asm_code_callback_aarch64 = {
  asm_code_aarch64_get_word_size,
  asm_code_aarch64_get_pointer_size,

  asm_code_aarch64_funccall_args_reg_number,
  asm_code_aarch64_funccall_args_stack_number,
  asm_code_aarch64_tmp_reg_number,
  asm_code_aarch64_tmp_stack_number,
  asm_code_aarch64_function_register_number,
  asm_code_aarch64_function_saveparam_number,

  asm_code_aarch64_stack_align_size,
  asm_code_aarch64_stack_correct_size,

  asm_code_aarch64_get_value,
  asm_code_aarch64_get_value_r1,
  asm_code_aarch64_get_address_stack,
  asm_code_aarch64_get_address,
  asm_code_aarch64_add_address,
  asm_code_aarch64_get_r1,
  asm_code_aarch64_set_r1,

  asm_code_aarch64_memory_load,
  asm_code_aarch64_memory_store,
  asm_code_aarch64_stack_load,
  asm_code_aarch64_stack_store,
  asm_code_aarch64_stack_load_r1,
  asm_code_aarch64_stack_store_r1,
  asm_code_aarch64_stack_expand,
  asm_code_aarch64_stack_reduce,

  asm_code_aarch64_funccall_reg_load,
  asm_code_aarch64_funccall_reg_store,
  asm_code_aarch64_tmp_reg_load,
  asm_code_aarch64_tmp_reg_save,
  asm_code_aarch64_tmp_reg_load_r1,
  asm_code_aarch64_tmp_reg_save_r1,

  asm_code_aarch64_sign_extension_char,
  asm_code_aarch64_sign_extension_uchar,
  asm_code_aarch64_sign_extension_short,
  asm_code_aarch64_sign_extension_ushort,
  asm_code_aarch64_sign_extension_int,
  asm_code_aarch64_sign_extension_uint,

  asm_code_aarch64_calc_inv,
  asm_code_aarch64_calc_minus,

  asm_code_aarch64_calc_op1,

  asm_code_aarch64_calc_add,
  asm_code_aarch64_calc_sub,
  asm_code_aarch64_calc_and,
  asm_code_aarch64_calc_or,
  asm_code_aarch64_calc_xor,

  asm_code_aarch64_calc_mul,
  asm_code_aarch64_calc_div,
  asm_code_aarch64_calc_mod,
  asm_code_aarch64_calc_llshift,
  asm_code_aarch64_calc_rashift,
  asm_code_aarch64_calc_rlshift,

  asm_code_aarch64_calc_op2,

  asm_code_aarch64_branch,
  asm_code_aarch64_branch_zero,
  asm_code_aarch64_branch_nzero,

  asm_code_aarch64_branch_cmp_eq,
  asm_code_aarch64_branch_cmp_ne,

  asm_code_aarch64_branch_cmp_lt,
  asm_code_aarch64_branch_cmp_gt,
  asm_code_aarch64_branch_cmp_le,
  asm_code_aarch64_branch_cmp_ge,
  asm_code_aarch64_branch_cmp_ult,
  asm_code_aarch64_branch_cmp_ugt,
  asm_code_aarch64_branch_cmp_ule,
  asm_code_aarch64_branch_cmp_uge,

  asm_code_aarch64_branch_cmp,

  asm_code_aarch64_function_call,
  asm_code_aarch64_function_call_pointer,

  asm_code_aarch64_function_start,
  asm_code_aarch64_function_end,
  asm_code_aarch64_function_register_save,
  asm_code_aarch64_function_register_load,
};
